#  Copyright 2022 The Matrix.org Foundation C.I.C.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from unittest import TestCase

from matrix_common.versionstring import get_distribution_version_string


class TestVersionString(TestCase):
    def test_our_own_version_string(self) -> None:
        """Sanity check that we get the version string for our own package.

        Check that it's a nonempty string.
        """
        version = get_distribution_version_string("matrix-common")
        self.assertIsInstance(version, str)
        self.assertTrue(version)



# Copyright 2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from unittest import TestCase

from matrix_common.regex import glob_to_regex, to_word_pattern


class GlobToRegexTestCase(TestCase):
    def test_literal_match(self) -> None:
        """Tests matching against a literal."""
        pattern = glob_to_regex("foobaz")
        self.assertRegex(
            "FoobaZ", pattern, "patterns should match and be case-insensitive"
        )
        self.assertNotRegex(
            "x foobaz", pattern, "pattern should not match at word boundaries"
        )

    def test_wildcard_match(self) -> None:
        """Tests matching with wildcards."""
        pattern = glob_to_regex("f?o*baz")

        self.assertRegex(
            "FoobarbaZ",
            pattern,
            "* should match string and pattern should be case-insensitive",
        )
        self.assertRegex("foobaz", pattern, "* should match 0 characters")
        self.assertNotRegex("fooxaz", pattern, "the character after * must match")
        self.assertNotRegex("fobbaz", pattern, "? should not match 0 characters")
        self.assertNotRegex("fiiobaz", pattern, "? should not match 2 characters")

    def test_multi_wildcard(self) -> None:
        """Tests matching with multiple wildcards in a row."""
        pattern = glob_to_regex("**baz")
        self.assertRegex("agsgsbaz", pattern, "** should match any string")
        self.assertRegex("baz", pattern, "** should match the empty string")
        self.assertEqual(pattern.pattern, r"\A(.{0,}baz)\Z")

        pattern = glob_to_regex("*?baz")
        self.assertRegex("agsgsbaz", pattern, "*? should match any string")
        self.assertRegex("abaz", pattern, "*? should match a single char")
        self.assertNotRegex("baz", pattern, "*? should not match the empty string")
        self.assertEqual(pattern.pattern, r"\A(.{1,}baz)\Z")

        pattern = glob_to_regex("a?*?*?baz")
        self.assertRegex("a g baz", pattern, "?*?*? should match 3 chars")
        self.assertNotRegex("a..baz", pattern, "?*?*? should not match 2 chars")
        self.assertRegex("a.gg.baz", pattern, "?*?*? should match 4 chars")
        self.assertEqual(pattern.pattern, r"\A(a.{3,}baz)\Z")

    def test_ignore_case(self) -> None:
        """Tests case sensitivity."""
        pattern = glob_to_regex("foobaz", ignore_case=False)
        self.assertEqual(pattern.flags & re.IGNORECASE, 0)

        pattern = glob_to_regex("foobaz", ignore_case=True)
        self.assertEqual(pattern.flags & re.IGNORECASE, re.IGNORECASE)


class WordPatternTestCase(TestCase):
    def test_whole_word(self) -> None:
        """Tests matching on whole words."""
        pattern = to_word_pattern("foo bar")

        self.assertRegex("foo bar", pattern)
        self.assertRegex(" foo bar ", pattern)
        self.assertRegex("baz foo bar baz", pattern)
        self.assertNotRegex("foo baré", pattern, "é should be seen as part of a word")
        self.assertNotRegex("bar foo", pattern, "Pattern should match words in order")

    def test_ends_with_non_letter(self) -> None:
        """Tests matching on whole words when the pattern ends with a space."""
        pattern = to_word_pattern("foo ")

        self.assertRegex(
            "foo bar",
            pattern,
            "Pattern should be able to end its match on a word boundary",
        )
        self.assertRegex(
            "foo ",
            pattern,
            "Pattern should be able to end its match at the end of a string",
        )
        self.assertRegex(
            "foo  ",
            pattern,
            "Pattern should be able to end its match anywhere",
        )



# Copyright 2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import List, Pattern

_WILDCARD_RUN = re.compile(r"([\?\*]+)")


def glob_to_regex(
    glob: str,
    *,
    word_boundary: bool = False,
    ignore_case: bool = True,
) -> Pattern[str]:
    """Converts a glob to a compiled regex object.

    Args:
        glob: pattern to match
        word_boundary: If `True`, the pattern will be allowed to match at word
            boundaries anywhere in the string. Otherwise, the pattern is
            anchored at the start and end of the string. When using this option,
            the pattern may match up to one extra non-word character on either
            side. The matching substring may be obtained from a capture group.
        ignore_case: If `True`, the pattern will be case-insensitive.
            Defaults to `True`.

    Returns:
        compiled regex pattern
    """
    # Patterns with wildcards must be simplified to avoid performance cliffs
    # - The glob `?**?**?` is equivalent to the glob `???*`
    # - The glob `???*` is equivalent to the regex `.{3,}`
    chunks: List[str] = []
    for chunk in _WILDCARD_RUN.split(glob):
        # No wildcards? re.escape()
        if not _WILDCARD_RUN.match(chunk):
            chunks.append(re.escape(chunk))
            continue

        # Wildcards? Simplify.
        question_marks = chunk.count("?")
        if "*" in chunk:
            chunks.append(".{%d,}" % (question_marks,))
        else:
            chunks.append(".{%d}" % (question_marks,))

    pattern = "".join(chunks)

    if word_boundary:
        pattern = to_word_pattern(pattern)
    else:
        # `\A` anchors at start of string, `\Z` at end of string
        # `\Z` is not the same as `$`! The latter will match the position before
        # a `\n` at the end of the string.
        pattern = rf"\A({pattern})\Z"

    return re.compile(pattern, re.IGNORECASE if ignore_case else 0)


def to_word_pattern(pattern: str) -> str:
    """Converts the given pattern to one that only matches on whole words.

    Adds word boundary characters to the start and end of a pattern to require that the
    match occur as a whole word.

    If the start or end of the pattern is a non-word character, then a word boundary is
    not required to precede or succeed it.

    A word boundary is considered to be the boundary between a word and non-word
    character. As such, the returned pattern is not appropriate for internationalized
    text search because there are languages which do not use spaces between words.

    Args:
        pattern: The pattern to wrap.

    Returns:
        A new pattern that only matches on whole words. The new pattern may match up to
        one extra non-word character on either side. The exact match is provided by a
        capture group.
    """
    # `^|\W` and `\W|$` handle the case where `pattern` starts or ends with a non-word
    # character.
    return rf"(?:^|\W|\b)({pattern})(?:\b|\W|$)"
